@@ -29,6 +29,7 @@ import bcrypt
 import asgineer
 import itemdb
 import pscript
+import iptools
 import timetagger
 from timetagger import config
 from timetagger.server import (
@@ -233,7 +234,8 @@ def load_credentials():
 
 
 def load_trusted_proxies():
-    return [s.strip() for s in config.proxy_auth_trusted.replace(";", ",").split(",")]
+    ips = [s.strip() for s in config.proxy_auth_trusted.replace(";", ",").split(",")]
+    return iptools.IpRangeList(*ips)
 
 
 CREDENTIALS = load_credentials()
@@ -241,4 +243,6 @@ TRUSTED_PROXIES = load_trusted_proxies()
 
 
 if __name__ == "__main__":
-    asgineer.run(main_handler, "uvicorn", config.bind, log_level="warning")
+    asgineer.run(
+        "timetagger.__main__:main_handler", "uvicorn", config.bind, log_level="warning"
+    )
