@@ -334,7 +334,6 @@ class SettingsStore(BaseStore):
         self.put(ob)
 
     def get_tag_info(self, tagz):
-
         info = {"targets": {}}
 
         # Load color the old way. We can remove this in about a year or so (2023)
@@ -373,6 +372,7 @@ class RecordStore(BaseStore):
 
     def __init__(self, datastore):
         self._datastore = datastore
+        self.put_count = 0  # Handy to detect new users
         self._items = {}  # key -> record
         self._running_records = {}  # Should be 0, 1, or occasionally maybe a few.
         self._heap = [{}]  # list of layers, each layer is binNr -> stats
@@ -446,6 +446,7 @@ class RecordStore(BaseStore):
 
             # Store the record in our flat dict
             self._items[key] = new_record
+            self.put_count += 1
 
             # Remove cur_record from bins in layer 0
             if cur_record is not None:
@@ -476,6 +477,18 @@ class RecordStore(BaseStore):
 
     def _update_bins(self, level, changed_bins):
         """Update bins of the given layer."""
+        # This uses a loop to avoid eaching the recursion depth limit
+        PSCRIPT_OVERLOAD = False  # noqa
+
+        while True:
+            result = self._update_bins_at_level(level, changed_bins)
+            if result is not None:
+                changed_bins = result
+                level += 1
+            else:
+                break
+
+    def _update_bins_at_level(self, level, changed_bins):
         PSCRIPT_OVERLOAD = False  # noqa
 
         # Init
@@ -524,10 +537,12 @@ class RecordStore(BaseStore):
             # Remove this level and all levels above
             while len(self._heap) > level:
                 self._heap.pop(-1)
+            return None
         elif len(heaplayer.keys()) == 1:
             # Remove all levels above here
             while len(self._heap) > level + 1:
                 self._heap.pop(-1)
+            return None
         else:
             # Bubble the changes up the heap
             changed_bins2 = {}
@@ -538,7 +553,7 @@ class RecordStore(BaseStore):
                 self._heap.append({})  # New layer, update all of it
                 for nr in heaplayer.keys():
                     changed_bins2[int(nr) // 2] = True
-            self._update_bins(level + 1, changed_bins2)
+            return changed_bins2
 
     def get_running_records(self):
         """Get a list of (copies of) the running records."""
@@ -849,6 +864,7 @@ class ConnectedDataStore(BaseDataStore):
         auth = self.get_auth()
         if not auth or auth.cantuse:
             self._set_state("error")
+            self.last_error = auth.cantuse or "Not authenticated"
             return
         # Get from local cache?
         if self._server_time == 0:
@@ -868,7 +884,6 @@ class ConnectedDataStore(BaseDataStore):
         await window.fetch(url, init)
 
     async def _push(self, kind, authtoken):
-
         # Take items, only proceed if nonempty
         items = self._to_push[kind]
         if len(items.keys()) == 0:
@@ -894,12 +909,33 @@ class ConnectedDataStore(BaseDataStore):
             # Put items back, but don't overwrite if the item was updated again.
             for key, item in items.items():
                 self._to_push[kind].setdefault(key, item)
-            self._set_state("error")  # is usually less bad than the fail's below
             text = await res.text()
-            console.warn(res.status + " (" + res.statusText + ") " + text)
-            # Also notify the user for 402 errors
-            if res.status == 402 and window.canvas:
-                window.canvas.notify_once(text)
+            self._set_state("error")  # is usually less bad than the fail's below
+            self.last_error = res.status + " (" + res.statusText + ")<br>" + text
+            # We have reserved status 402 for messages that we want to notify the user of.
+            # Note that the 402 status code is defined as "payment required", which
+            # may not be what we mean here, so we replace the statusText.
+            if res.status == 402:
+                statusText = "notification"
+                i = text.find("|")
+                if i >= 0 and i < 16:
+                    statusText, _, text = text.partition("|")
+                    text = text.lstrip()
+                self.last_error = res.status + " (" + statusText + ")<br>" + text
+                # We also handle the case where the server forbids
+                # pushes, i.e. the account is read-only, so that the
+                # dialogs can disable buttons accordingly. We use CSS
+                # to disable all actionbuttons. And an attribbute to
+                # cover some specific edits in front.py. Note that the
+                # user can work around this thing and make changes, but
+                # these changes cannot be pushed to the server.
+                if statusText.lower() == "readonly":
+                    window.document.body.classList.add("is_read_only")
+                    self.is_read_only = True
+                # Notify user
+                if window.canvas:
+                    window.canvas.notify_once(text)
+            console.warn(self.last_error)
 
         else:
             # Success, but it can still mean that some records failed. In this
@@ -911,10 +947,10 @@ class ConnectedDataStore(BaseDataStore):
                 self[kind]._drop(key)
             for err in d.errors:
                 self._set_state("warning")
-                console.warn(f"Server dropped a {kind}: {err}")
+                self.last_error = f"Server dropped a {kind}: {err}"
+                console.warn(self.last_error)
 
     async def _pull(self, authtoken):
-
         # Fetch and wait for response
         url = tools.build_api_url("updates?since=" + self._server_time)
         init = dict(method="GET", headers={"authtoken": authtoken})
@@ -928,8 +964,9 @@ class ConnectedDataStore(BaseDataStore):
         # Process response
         if res.status != 200:
             text = await res.text()
-            console.warn(res.status + " (" + res.statusText + ") " + text)
             self._set_state("error")  # E.g. Wifi or server down, or 500
+            self.last_error = res.status + " (" + res.statusText + ")<br>" + text
+            console.warn(self.last_error)
             if res.status == 401:
                 # Our token is probably expired. There may be local
                 # changes that have not yet been pushed, which would
@@ -955,12 +992,14 @@ class ConnectedDataStore(BaseDataStore):
                     self.settings._put_received(*ob.settings)
                 except Exception as err:
                     self._set_state("warning")
+                    self.last_error = err
                     console.error(err)
                     window.alert("Sync error (settings), see dev console for details.")
                 try:
                     self.records._put_received(*ob.records)
                 except Exception as err:
                     self._set_state("warning")
+                    self.last_error = err
                     console.error(err)
                     window.alert("Sync error (records), see dev console for details.")
 
@@ -1003,6 +1042,15 @@ class DemoDataStore(BaseDataStore):
         self._create_tags()
         self._create_one_year_of_data(self._years.pop(-1))
 
+        # Add many small records, for testing
+        # ds = "hello #there"
+        # t = dt.now()
+        # for i in range(20):
+        #     t2 = t + 220
+        #     record = self.records.create(t, t2, ds)
+        #     t = t2
+        #     self.records._put_received(record)
+
     async def _sync(self):
         """Emulate a sync action."""
         for kind in ["settings", "records"]:
@@ -1030,7 +1078,6 @@ class DemoDataStore(BaseDataStore):
             self.settings.set_tag_info(tag, {"color": color})
 
     def _create_tags(self):
-
         self._tag_groups1 = [
             ["#admin", "#reading"],
             ["#admin", "#traveling"],
