@@ -56,7 +56,9 @@ def fit_font_size(ctx, available_width, font, text, maxsize=100):
     # Takes 2-5 iters, smaller available_width -> faster iteration
     size = maxsize
     width = available_width + 2
-    while width > available_width and size > 4:
+    iter = 0  # failsafe
+    while iter < 9 and width > available_width and size > 4:
+        iter += 1
         new_size = int(1.1 * size * available_width / width)
         size = new_size if new_size < size else size - 1
         ctx.font = str(size) + "px " + font
@@ -104,7 +106,6 @@ def color_from_name(name):
 
 
 def create_palettes():
-
     # The Github color palette, consisting of 8 strong colors and 8 lighter variants.
     # The difference between these variants is pretty big.
     # gh_colors = (
@@ -461,20 +462,10 @@ def positions_mean_and_std(positions):
     return avg_pos, std_pos
 
 
-def get_pixel_ratio(ctx):
+def get_pixel_ratio():
     """Get the ratio of logical pixel to screen pixel."""
     PSCRIPT_OVERLOAD = False  # noqa
-
-    dpr = window.devicePixelRatio or 1
-    bsr = (
-        ctx.webkitBackingStorePixelRatio
-        or ctx.mozBackingStorePixelRatio
-        or ctx.msBackingStorePixelRatio
-        or ctx.oBackingStorePixelRatio
-        or ctx.backingStorePixelRatio
-        or 1
-    )
-    return dpr / bsr
+    return window.devicePixelRatio or 1
 
 
 def create_pointer_event(node, e):
@@ -611,7 +602,7 @@ class SimpleSettings:
     Key-value pairs are stored in a cache for fast getters.
     The actual storage can be:
     * None / session only: unknown keys are not stored across sessions.
-    * Local storagee: a prefefined set of keys are stored in local storaged,
+    * Local storage: a prefefined set of keys are stored in local storaged,
       which means that they are device-specific.
     * Synced: synced to the server using the settings store.
     """
@@ -621,17 +612,22 @@ class SimpleSettings:
         self._local_keys = {
             "darkmode": 1,
             "width_mode": "auto",
+            "notifications": False,
             "pomodoro_enabled": False,
             "report_grouping": "date",
+            "report_groupperiod": "none",
             "report_hidesecondary": False,
-            "report_hourdecimals": False,
+            "report_format": "hm",
             "report_showrecords": True,
         }
         self._synced_keys = {
             "first_day_of_week": 1,
+            "workdays": 0,
+            "date_repr": "dd-mm-yyyy",
             "time_repr": "auto",
             "duration_repr": "hms",
             "today_snap_offset": "",
+            "today_end_offset": "",
             "show_stopwatch": True,
         }
         # The data store for synced source
@@ -739,7 +735,6 @@ class BaseCanvas:
         self._draw_tick()
 
     def _init_events(self):
-
         # Disable context menu so we can handle RMB clicks
         # Firefox is particularly stuborn with Shift+RMB, and RMB dbl click
         for ev_name in ("contextmenu", "click", "dblclick"):
@@ -777,8 +772,9 @@ class BaseCanvas:
         self.node.addEventListener("touchmove", self._on_js_touch_event, 0)
 
         # Keep track of window size
-        window.addEventListener("resize", self._on_js_resize_event, False)
-        window.setTimeout(self._on_js_resize_event, 10)
+        # window.addEventListener("resize", self._on_js_resize_event, False)
+        self._resize_observer = window.ResizeObserver(self._on_resize_observer)
+        self._resize_observer.observe(self.node)
 
     def _prevent_default_event(self, e):
         """Prevent the default action of an event unless all modifier
@@ -839,21 +835,55 @@ class BaseCanvas:
         }.get(e.type[5:])
         self.on_pointer(ev)
 
-    def _on_js_resize_event(self):
-        """Ensure that the canvas has the correct size and dpi."""
-        ctx = self.node.getContext("2d")
-        self.pixel_ratio = get_pixel_ratio(ctx)
+    def _on_resize_observer(self, entries):
+        # The resize observer allows us to get the extact size of the element
+        # in physical pixels, something we cannot do with a normal resize event.
+        # See https://web.dev/articles/device-pixel-content-box
+        #
+        # On Firefox I've seen the app hang due to window resizing. I have not
+        # been able to find a singular reason, but it seems important that we
+        # update the canvas physical size (i.e. call _apply_new_size) directly.
+        # It also looks like drawing stuff outside a requested animation frame
+        # is dangerous. See https://github.com/almarklein/timetagger/pull/418
+        entry = entries.find(lambda entry: entry.target is self.node)
+        if entry.devicePixelContentBoxSize:
+            # Best if we have the physical pixels ...
+            psize = [
+                entry.devicePixelContentBoxSize[0].inlineSize,
+                entry.devicePixelContentBoxSize[0].blockSize,
+            ]
+        else:
+            # ... but not all browsers support that (see issue #423) ...
+            if entry.contentBoxSize:
+                lsize = [
+                    entry.contentBoxSize[0].inlineSize,
+                    entry.contentBoxSize[0].blockSize,
+                ]
+            else:  # even more backward compat
+                lsize = [entry.contentRect.width, entry.contentRect.height]
+            ratio = get_pixel_ratio()
+            psize = Math.floor(lsize[0] * ratio), Math.floor(lsize[1] * ratio)
 
-        # A line-width of 2 is great to have crisp images. For uneven line widths
-        # one needs to offset 0.5 * pixel_ratio. But, that line-width must be
-        # snapped to a width matching the pixel_ratio! pfew!
-        self.grid_linewidth2 = min(self.pixel_ratio, self.grid_round(1)) * 2
-
-        self.w, self.h = self.node.clientWidth, self.node.clientHeight
-        self.node.width = self.w * self.pixel_ratio
-        self.node.height = self.h * self.pixel_ratio
+        self._apply_new_size(psize)
+        # Calling self._draw() avoids flicker, but I still saw occasional hangups,
+        # so let's accept the flicker and use update() ...
         self.update()
-        self.on_resize(True)  # draw asap to avoid flicker
+
+    def _apply_new_size(self, psize):
+        # This is called JIT right before a draw, when a resize has happened
+
+        # Get and store pixel ratio
+        self.pixel_ratio = ratio = get_pixel_ratio()
+        # Use that to obtain the real number of logical pixels
+        lsize = psize[0] / ratio, psize[1] / ratio
+        # Apply
+        self.node.width = psize[0]
+        self.node.height = psize[1]
+        self.w, self.h = lsize[0], lsize[1]
+        # Calculate linewidth param
+        self.grid_linewidth2 = min(self.pixel_ratio, self.grid_round(1)) * 2
+        # Update
+        self.on_resize()
 
     def grid_round(self, x):
         """Round a value to the screen pixel grid."""
@@ -873,8 +903,8 @@ class BaseCanvas:
 
     def update(self, asap=True):
         """Schedule an update."""
-        # The extra setTimeout is to make sure that there is time for the
-        # browser to process events (like scrolling).
+        # The optional extra setTimeout is to make sure that there is
+        # time for the browser to process events (like scrolling).
         if not self._pending_draw:
             self._pending_draw = True
             if asap:
@@ -884,22 +914,25 @@ class BaseCanvas:
 
     def _draw(self):
         """The entry draw function, called by the browser."""
+
+        # Reset flag
         self._pending_draw = False
+
+        # Is the element even visible
         if self.node.style.display == "none":
             return  # Hidden
-        elif self.w <= 0 or self.h <= 0:
+        if self.w <= 0 or self.h <= 0:
             return  # Probably still initializing
 
-        ctx = self.node.getContext("2d")
+        # Some bookkeeping
+        self._tooltips.clear()
 
-        # Prepare hidpi mode for canvas  (flush state just in case)
-        for i in range(4):
-            ctx.restore()
+        # Prepare context
+        ctx = self.node.getContext("2d")
+        ctx.restore()  # undo last
         ctx.save()
         ctx.scale(self.pixel_ratio, self.pixel_ratio)
 
-        self._tooltips.clear()
-
         # Draw
         self.on_draw(ctx)
 
@@ -908,7 +941,7 @@ class BaseCanvas:
         x, y = ev.pos
         # Only on mouse - it's anoying for touch
         is_mouse = bool(ev.touches["-1"])
-        if not is_mouse:
+        if not (is_mouse and looks_like_desktop()):
             self._tooltipdiv.style.display = "none"
             return
         # Get tooltip object - if text is None it means no tooltip
